(define-module (language tree-il eval)
  #:use-module (system base syntax)
  #:use-module (language tree-il)
  #:use-module (system base pmatch)
  #:use-module (srfi srfi-26)
  #:use-module (srfi srfi-11)
  #:use-module (srfi srfi-1)
  #:use-module (ice-9 format)
  #:export (eval-tree-il))

(define guile-module (resolve-module '(guile)))

(define (eval1 x e)
  (record-case x
    ((<void>)
     (if #f #f))
    ((<const> src exp)
     exp)
    ((<primitive-ref> src name)
     (module-ref guile-module name))
    ((<lexical-ref> src name gensym)
     (cond
      ((assq gensym e) => cdr)
      (else (error "Unbound variable" name))))
    ((<lexical-set> src name gensym exp)
     (cond
      ((assq gensym e)
       => (lambda (binding)
            (set-cdr! binding (eval1 exp e))
            (if #f #f)))
      (else (error "Unbound variable" name))))
    ((<module-ref> src mod name public?)
     (module-ref (resolve-module mod) name))
    ((<module-set> src mod name public? exp)
     (module-set! (resolve-module mod) name (eval1 exp e)))
    ((<toplevel-ref> src name)
     (module-ref (current-module) name))
    ((<toplevel-set> src name exp)
     (module-set! (current-module) name (eval1 exp e)))
    ((<toplevel-define> src name exp)
     (module-define! (current-module) name (eval1 exp e)))
    ((<conditional> src test consequent alternate)
     (if (eval1 test e)
         (eval1 consequent e)
         (eval1 alternate e)))
    ((<call> src proc args)
     (apply (eval1 proc e) (map (cut eval1 <> e) args)))
    ((<primcall> src name args)
     (apply (module-ref guile-module name)
            (map (lambda (x) (eval1 x e)) args)))
    ((<seq> src head tail)
     (begin (eval1 head e) (eval1 tail e)))
    ((<lambda> src meta body)
     (let ((v (eval1 body e)))
       (set-procedure-properties! v meta)
       v))
    ((<lambda-case> req opt rest kw inits gensyms alternate body)
     (lambda args
       (define (fail)
         (if alternate
             (eval1 alternate e)
             (error "Wrong number of arguments")))
       (define (parse-req args req gensyms e)
         (if (pair? req)
             (if (pair? args)
                 (parse-req (cdr args)
                            (cdr req)
                            (cdr gensyms)
                            (acons (car gensyms) (car args) e))
                 (fail))
             (parse-opt args opt inits gensyms e)))
       (define (parse-opt args opt inits gensyms e)
         (if (pair? opt)
             (if (pair? args)
                 (parse-opt (cdr args)
                            (cdr opt)
                            (cdr inits)
                            (cdr gensyms)
                            (acons (car gensyms) (car args) e))
                 (parse-opt args
                            (cdr opt)
                            (cdr inits)
                            (cdr gensyms)
                            (acons (car gensyms) (eval1 (car inits) e) e)))
             (parse-rest args rest inits gensyms e)))
       (define (parse-rest args rest inits gensyms e)
         (if rest
             (parse-kw args kw inits (cdr gensyms) (acons (car gensyms) args e))
             (parse-kw args rest inits gensyms e)))
       (define (parse-kw args kw inits gensyms e)
         (if kw
             (parse-kw1 args (cdr kw) (car kw) (map cons gensyms inits) e)
             (finish e)))
       (define (parse-kw1 args kw aok? g+i e)
         (if (pair? args)
             (if (pair? (cdr args))
                 (let ((arg-kw (car args))
                       (arg-val (cadr args))
                       (next (cddr args)))
                   (cond
                    ((assq arg-kw kw)
                     => (lambda (kw+name+var)
                          (let ((gensym (caddr kw+name+var)))
                            (parse-kw1 next
                                       kw
                                       aok?
                                       (alist-delete gensym g+i eq?)
                                       (acons gensym arg-val e)))))
                    (aok?
                     (parse-kw1 next kw aok? g+i e))
                    (else
                     (fail))))
                 (fail))
             (if g+i
                 (parse-kw1 args
                            kw
                            aok?
                            (cdr g+i)
                            (acons (caar g+i) (eval1 (cdar g+i) e) e))
                 (finish e))))
       (define (finish e)
         (eval1 body e))
       (parse-req args req gensyms e)))
    ((<let> src names gensyms vals body)
     (eval1 body
            (append (map cons gensyms (map (cut eval1 <> e) vals)) e)))
    ((<letrec> in-order? src names gensyms vals body)
     (let ((e0 (append (map (cut cons <> #f) gensyms) e)))
       (let loop ((gensyms gensyms)
                  (vals vals)
                  (e e0))
         (if (pair? gensyms)
             (loop (cdr gensyms)
                   (cdr vals)
                   (assq-set! e (car gensyms) (eval1 (car vals) e0)))
             (eval1 body e)))))
    ((<prompt> tag body handler)
     (call-with-prompt (eval1 tag e)
                       (lambda () (eval1 body e))
                       (eval1 handler e)))
    ((<abort> tag args tail)
     (apply abort-to-prompt
            (eval1 tag e)
            (append (map (cut eval1 <> e) args) (eval1 tail e))))))

(define (eval-tree-il x)
  (eval1 x '()))
